/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <dlog.h>

#include <memory>

#include "bundle_cpp_implementation.h"
#include "bundle_cpp.h"
#include "bundle_internal.h"


#ifdef LOG_TAG
#undef LOG_TAG
#endif

#define LOG_TAG "BUNDLE"

namespace tizen_base {
Bundle::Impl::Impl(Bundle* parent, bool copy, bool own)
  : copy_(copy), own_(own), parent_(parent) {
}

Bundle::Impl::Impl(Bundle* parent) : parent_(parent) {
}

Bundle::Impl::~Impl() = default;

Bundle::Bundle()
  : impl_(new Impl(this)) {
  impl_->handle_ = bundle_create();
  if (impl_->handle_ == nullptr)
    throw std::bad_alloc();
}

Bundle::Bundle(BundleRaw raw)
  : impl_(new Impl(this)) {
  impl_->handle_ = bundle_decode(raw.first.get(), raw.second);
  if (impl_->handle_ == nullptr)
    throw std::bad_alloc();
}

Bundle::Bundle(const std::string& raw)
  : impl_(new Impl(this)) {
  impl_->handle_ = bundle_decode(reinterpret_cast<const bundle_raw*>(
      raw.c_str()), raw.length());
  if (impl_->handle_ == nullptr)
    throw std::bad_alloc();
}

Bundle::Bundle(bundle* b, bool copy, bool own)
  : impl_(new Impl(this, copy, own)) {
  if (b == nullptr)
    throw std::invalid_argument("b cannot be null");

  if (!impl_->copy_) {
    impl_->handle_ = b;
  } else {
    impl_->handle_ = bundle_dup(b);
    if (impl_->handle_ == nullptr)
      throw std::bad_alloc();
  }
}

Bundle::~Bundle() {
  if (impl_->handle_ && (impl_->own_ || impl_->copy_))
    bundle_free(impl_->handle_);
}

Bundle::Bundle(const Bundle& b)
  : impl_(new Impl(this)) {
  impl_->handle_ = bundle_dup(b.impl_->handle_);
  if (impl_->handle_ == nullptr)
    throw std::bad_alloc();
}

Bundle::KeyInfo::KeyInfo(const bundle_keyval_t* handle, std::string name)
  : impl_(new Impl(this, handle, std::move(name))) {
}

Bundle::KeyInfo::KeyInfo()
  : impl_(new Impl(this)) {
}

Bundle::KeyInfo::~KeyInfo() {
}

Bundle::KeyInfo::Impl::~Impl() = default;
Bundle::KeyInfo::Impl::Impl(Bundle::KeyInfo* parent,
                            const bundle_keyval_t* handle,
                            std::string name)
  : handle_(handle), name_(name), parent_(parent) {
}

Bundle::KeyInfo::Impl::Impl(Bundle::KeyInfo* parent) : parent_(parent) {
}

Bundle::KeyInfo::KeyInfo(const KeyInfo& k)
    : impl_(new Impl(this)) {
  impl_->handle_ = bundle_keyval_dup(k.impl_->handle_);
  impl_->name_ = k.impl_->name_;
  if (impl_->handle_ == nullptr)
    throw std::bad_alloc();
}

Bundle::KeyInfo& Bundle::KeyInfo::operator = (const Bundle::KeyInfo& k) {
  if (this != &k) {
    impl_->handle_ = bundle_keyval_dup(k.impl_->handle_);
    impl_->name_ = k.impl_->name_;
    if (impl_->handle_ == nullptr)
      throw std::bad_alloc();
  }
  return *this;
}

Bundle::KeyInfo::KeyInfo(Bundle::KeyInfo&& k) noexcept {
  impl_ = std::unique_ptr<Impl>(new Impl(this));
  impl_->handle_ = k.impl_->handle_;
  impl_->name_ = k.impl_->name_;
  k.impl_->handle_ = nullptr;
  k.impl_->name_ = "";
}

Bundle::KeyInfo& Bundle::KeyInfo::operator = (Bundle::KeyInfo&& k) noexcept {
  if (this != &k) {
    impl_->handle_ = k.impl_->handle_;
    impl_->name_ = k.impl_->name_;
    k.impl_->handle_ = nullptr;
    k.impl_->name_ = "";
  }
  return *this;
}

bundle_type Bundle::KeyInfo::GetType() const {
  return static_cast<bundle_type>(
      bundle_keyval_get_type(const_cast<bundle_keyval_t*>(impl_->handle_)));
}

bool Bundle::KeyInfo::IsArray() const {
  return bundle_keyval_type_is_array(const_cast<bundle_keyval_t*>(
      impl_->handle_));
}

const std::string& Bundle::KeyInfo::GetName() const {
  return impl_->name_;
}

Bundle& Bundle::operator = (const Bundle& b) {
  if (this != &b) {
    impl_->handle_ = bundle_dup(b.impl_->handle_);
    if (impl_->handle_ == nullptr)
      throw std::bad_alloc();
  }
  return *this;
}

Bundle::Bundle(Bundle&& b) noexcept {
  impl_ = std::unique_ptr<Impl>(new Impl(this));
  impl_->handle_ = b.impl_->handle_;
  b.impl_->handle_ = nullptr;
}

Bundle& Bundle::operator = (Bundle&& b) noexcept {
  if (this != &b) {
    impl_->handle_ = b.impl_->handle_;
    b.impl_->handle_ = nullptr;
  }
  return *this;
}

std::vector<Bundle::KeyInfo> Bundle::GetKeys() {
  std::vector<Bundle::KeyInfo> v;

  bundle_foreach(impl_->handle_, [](const char *key, const int type,
      const bundle_keyval_t *kv, void *user_data) {
        auto* v = static_cast<std::vector<KeyInfo>*>(user_data);
        v->emplace_back(kv, key);
      }, &v);

  return v;
}

int Bundle::Add(const std::string& key, const std::string& val) {
  int ret = bundle_add_str(impl_->handle_, key.c_str(), val.c_str());
  if (ret != BUNDLE_ERROR_NONE)
    LOGE("Add fail key(%s), val(%s), ret(%d)", key.c_str(), val.c_str(), ret);
  return ret;
}

int Bundle::Add(const std::string& key, const std::vector<std::string>& val) {
  std::vector<const char*> v;
  for (auto& i : val) {
    v.push_back(i.c_str());
  }

  int ret = bundle_add_str_array(
    impl_->handle_, key.c_str(), v.data(), v.size());
  if (ret != BUNDLE_ERROR_NONE)
    LOGE("Add fail key(%s), ret(%d)", key.c_str(), ret);

  return ret;
}

int Bundle::Add(const std::string& key, const std::vector<unsigned char>& val) {
  int ret = bundle_add_byte(impl_->handle_, key.c_str(), val.data(), val.size());
  if (ret != BUNDLE_ERROR_NONE)
    LOGE("Add fail key(%s), ret(%d)", key.c_str(), ret);
  return ret;
}

int Bundle::Delete(const std::string& key) {
  int ret = bundle_del(impl_->handle_, key.c_str());
  if (ret != BUNDLE_ERROR_NONE)
    LOGE("Add fail key(%s), ret(%d)", key.c_str(), ret);
  return ret;
}

std::string Bundle::GetString(const std::string& key) const {
  char* str = nullptr;
  bundle_get_str(impl_->handle_, key.c_str(), &str);

  if (!str)
    return "";

  return std::string(str);
}

std::vector<std::string> Bundle::GetStringArray(const std::string& key) const {
  std::vector<std::string> v;

  const char** str_array = nullptr;
  int len = 0;

  str_array = bundle_get_str_array(impl_->handle_, key.c_str(), &len);

  for (int i = 0; i < len; i++) {
    v.emplace_back(str_array[i]);
  }

  return v;
}

std::vector<unsigned char> Bundle::GetByte(const std::string& key) const {
  size_t size;
  unsigned char* bytes = nullptr;
  bundle_get_byte(impl_->handle_, key.c_str(),
      reinterpret_cast<void**>(&bytes), &size);
  return std::vector<unsigned char>(bytes, bytes + size);
}

Bundle::BundleRaw Bundle::ToRaw() {
  bundle_raw* raw = nullptr;
  int len = 0;
  int ret = bundle_encode(impl_->handle_, &raw, &len);
  if (raw == nullptr) {
    LOGE("Fail to encode data (%d)", ret);
    throw std::bad_alloc();
  }

  return BundleRaw(
      std::unique_ptr<bundle_raw, decltype(std::free)*>(raw, std::free), len);
}

int Bundle::GetCount() const {
  return bundle_get_count(impl_->handle_);
}

bundle_type Bundle::GetType(const std::string& key) const {
  return static_cast<bundle_type>(bundle_get_type(impl_->handle_, key.c_str()));
}

bundle* Bundle::GetHandle() const {
  return impl_->handle_;
}

bundle* Bundle::Detach() {
  auto* h = impl_->handle_;
  impl_->handle_ = nullptr;
  return h;
}

}  // namespace tizen_base
